from calendar import c
from concurrent.futures import ThreadPoolExecutor
import asyncio
import threading
from koil.vars import *


class KoiledExecutor(ThreadPoolExecutor):
    def submit(self, fn, *args, cancel_event=None, **kwargs):
        copy_loop = asyncio.get_event_loop()
        copy_loop.run_in_executor()

        assert (
            copy_loop.is_running()
        ), "Loop is not running. You shouldn't be using this"

        def wrap(*args, **kwargs):
            current_loop.set(copy_loop)
            current_cancel_event.set(cancel_event)
            t = fn(*args, **kwargs)
            current_loop.set(None)
            current_cancel_event.set(None)
            return t

        return super().submit(wrap, *args, **kwargs)

    async def asubmit(self, fn, *args, **kwargs):

        cancel_event = threading.Event()

        try:
            return await asyncio.wrap_future(
                super().submit(fn, *args, **kwargs, cancel_event=cancel_event)
            )
        except asyncio.CancelledError:
            cancel_event.set()
            raise



import asyncio
from asyncio.futures import Future
import contextvars
import inspect
from typing import Callable, Generic, TypeVar
from qtpy import QtGui
from qtpy.QtCore import QObject, Signal, QThread
import uuid
import logging
from koil.koil import Koil
from qtpy import QtWidgets
from qtpy import QtCore
from concurrent import futures
from typing_extensions import ParamSpec, final
from koil.task import KoilGeneratorTask, KoilTask


logger = logging.getLogger(__name__)


Reference = str


class UnconnectedSignalError(Exception):
    pass


def get_receiver_length(qobject, qsignal, callstring):
    try:
        return qobject.receivers(qsignal)
    except:
        return qobject.receivers(callstring)


class QtFuture:
    def __init__(self):
        self.loop = asyncio.get_event_loop()
        self.aiofuture = asyncio.Future()
        self.iscancelled = False

    def _set_cancelled(self):
        """WIll be called by the asyncio loop"""
        self.iscancelled = True

    def resolve(self, *args):
        ctx = contextvars.copy_context()
        self.loop.call_soon_threadsafe(self.aiofuture.set_result, (ctx,) + args)

    def reject(self, exp: Exception):
        self.loop.call_soon_threadsafe(self.aiofuture.set_exception, exp)


T = TypeVar("T")
P = ParamSpec("P")


class QtCoro(QtCore.QObject, Generic[T, P]):
    called = QtCore.Signal(QtFuture, tuple, dict, object)
    cancelled = QtCore.Signal(str)

    def __init__(
        self,
        coro: Callable[P, T],
        *args: P.args,
        autoresolve=False,
        use_context=True,
        **kwargs: P.kwargs,
    ):
        super().__init__(*args, **kwargs)
        assert not inspect.iscoroutinefunction(
            coro
        ), "This should not be a coroutine, but a normal qt slot with the first parameter being a qtfuture"
        self.coro = coro
        self.called.connect(self.on_called)
        self.autoresolve = autoresolve
        self.use_context = use_context

    def on_called(self, future, args, kwargs, ctx):

        try:
            if self.use_context:
                for ctx, value in ctx.items():
                    ctx.set(value)

            x = self.coro(future, *args, **kwargs)
            if self.autoresolve:
                future.resolve(x)
        except Exception as e:
            logger.error(f"Error in Qt Coro {self.coro}", exc_info=True)
            future.reject(e)

    async def acall(self, *args: P.args, timeout=None, **kwargs: P.kwargs):
        qtfuture = QtFuture()
        ctx = contextvars.copy_context()
        self.called.emit(qtfuture, args, kwargs, ctx)
        try:
            if timeout:
                context, x = await asyncio.wait_for(qtfuture.aiofuture, timeout=timeout)
            else:
                context, x = await qtfuture.aiofuture

            for ctx, value in context.items():
                ctx.set(value)

            return x

        except asyncio.CancelledError:
            qtfuture._set_cancelled()
            raise


class QtTask(KoilTask, QtCore.QObject):
    errored = QtCore.Signal(Exception)
    cancelled = QtCore.Signal()
    returned = QtCore.Signal(object)
    _returnedwithoutcontext = QtCore.Signal(object, object)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._returnedwithoutcontext.connect(self.on_returnedwithoutcontext)

    def on_returnedwithoutcontext(self, obj, context):
        for ctx, value in context.items():
            ctx.set(value)

        self.returned.emit(obj)

    async def wrapped_future(self, args, kwargs, context):
        try:
            for ctx, value in context.items():
                ctx.set(value)

            args = self.args + args
            kwargs = {**self.kwargs, **kwargs}
            i = await self.coro(*args, **kwargs)
            newcontext = contextvars.copy_context()
            self._returnedwithoutcontext.emit(i, newcontext)
        except Exception as e:
            logger.error(
                f"Error in task of coroutine {self.coro.__name__}", exc_info=True
            )
            self.errored.emit(e)
        except asyncio.CancelledError:
            self.cancelled.emit()

    def run(self, *args, **kwargs):

        ctxs = contextvars.copy_context()
        future = asyncio.run_coroutine_threadsafe(
            self.wrapped_future(args, kwargs, ctxs), self.loop
        )
        return future


class QtGeneratorTask(KoilGeneratorTask, QtCore.QObject):
    errored = QtCore.Signal(Exception)
    cancelled = QtCore.Signal()
    yielded = QtCore.Signal(object)
    done = QtCore.Signal(object)
    _yieldedwithoutcontext = QtCore.Signal(object, object)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._yieldedwithoutcontext.connect(self.on_yieldedwithoutcontext)

    def on_yieldedwithoutcontext(self, x, context):
        for ctx, value in context.items():
            ctx.set(value)

        self.yielded.emit(x)

    async def wrapped_future(self, args, kwargs, ctxs):
        print("osinosinsoi")
        for ctx, value in ctxs.items():
            ctx.set(value)

        try:
            args = self.args + args
            kwargs = {**self.kwargs, **kwargs}
            print("nanana")
            async for i in self.iterator(*args, **kwargs):
                print(i)

                newcontext = contextvars.copy_context()
                self._yieldedwithoutcontext.emit(i, newcontext)
        except Exception as e:
            self.errored.emit(e)
        except asyncio.CancelledError:
            self.cancelled.emit()

    def run(self, *args, **kwargs):
        ctxs = contextvars.copy_context()
        print("Hallo")

        future = asyncio.run_coroutine_threadsafe(
            self.wrapped_future(args, kwargs, ctxs), self.loop
        )
        return future


class QtKoil(Koil, QtCore.QObject):
    def __init__(
        self,
        *args,
        disconnect_on_close=True,
        autoconnect=True,
        parent=None,
        uvify=False,
        **kwargs,
    ) -> None:
        super().__init__(
            *args, **kwargs, uvify=False, task_class=QtTask, gen_class=QtGeneratorTask
        )
        self.disconnect_on_close = disconnect_on_close
        if autoconnect:
            self.connect()

    def close(self):
        self.__exit__(None, None, None)

    def connect(self):
        ap_instance = QtWidgets.QApplication.instance()
        if ap_instance is None:
            raise NotImplementedError("Qt Application not found")
        if self.disconnect_on_close:
            ap_instance.lastWindowClosed.connect(self.close)
        return self.__enter__()



import contextvars
import logging
import asyncio
import threading
import time
from typing import AsyncIterator, Awaitable, Callable, Coroutine, Generic, TypeVar
from koil.vars import current_cancel_event, current_loop
import inspect
from .utils import run_threaded_with_context
from typing_extensions import ParamSpec, final

logger = logging.getLogger(__name__)

T = TypeVar("T")
P = ParamSpec("P")


class KoilTask(Generic[T, P]):
    """Will run the giving coroutine in the thread of the loop
    and return the result

    This is a thin wrapper around concurrent.futures api, but with
    additional context support.

    Args:
        Generic (_type_): _description_
    """

    def __init__(
        self,
        coro: Callable[P, Awaitable[T]],
        *args,
        preset_args=(),
        preset_kwargs={},
        log_errors=True,
        loop=None,
        bypass_test=False,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.coro = coro
        self.args = preset_args
        self.kwargs = preset_kwargs
        self.loop = loop or current_loop.get()
        assert self.loop, "No koiled Loop found"
        self.task = None
        self.future = None
        if not bypass_test:
            assert self.loop.is_running(), "Loop is not running"
            assert not self.loop.is_closed(), "Loop is closed"
            assert inspect.iscoroutinefunction(coro), "Task is not a coroutine"

    def run(self, *args: P.args, **kwargs: P.kwargs):
        assert self.future is None, "Task is already running"
        args = self.args + args
        kwargs = {**self.kwargs, **kwargs}
        self.cancel_event = threading.Event()
        self.future = run_threaded_with_context(
            self.coro(*args, **kwargs), self.loop, self.cancel_event
        )

        return self

    def done(self):
        assert self.future, "Task was never run! Please run task before"
        return self.future.done()

    def cancel(self):
        assert not self.future.done(), "Task was never run! Please run task before"
        self.cancel_event.set()

    def result(self) -> T:
        assert self.future, "Task was never run! Please run task before"

        res, context = self.future.result()

        for ctx, value in context.items():
            ctx.set(value)

        return res


class KoilGeneratorTask(Generic[P, T]):
    def __init__(
        self,
        iterator: Callable[P, AsyncIterator[T]],
        args=(),
        kwargs={},
        loop=None,
        bypass_test=False,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.iterator = iterator
        self.args = args
        self.kwargs = kwargs
        self.loop = loop or current_loop.get()
        self.task = None
        self._task_done = False
        self._buffer = []
        self._latest_context = None
        if not bypass_test:
            assert self.loop.is_running(), "Loop is not running"
            assert not self.loop.is_closed(), "Loop is closed"
            assert inspect.isasyncgenfunction(iterator), "Task is not a async iterator"

    def run(self, *args: P.args, **kwargs: P.kwargs):
        ait = self.iterator(*self.args, **self.kwargs).__aiter__()
        res = [False, False]
        cancel_event = current_cancel_event.get()

        async def next_on_ait():
            try:
                try:
                    obj = await ait.__anext__()
                    return [False, obj]
                except StopAsyncIteration:
                    return [True, None]
            except asyncio.CancelledError as e:
                return [False, e]

        raise NotImplementedError("No design decision was taken")

    def done(self):
        return self._task_done

    def result(self):
        if not self._task_done:
            raise Exception("Task is not done yet")

        self._buffer

        raise NotImplementedError("This is a generator that yields results")



import contextvars

from koil.errors import ThreadCancelledError


current_koil = contextvars.ContextVar("current_koil", default=None)
current_loop = contextvars.ContextVar("current_loop", default=None)
current_taskclass = contextvars.ContextVar("current_taskclass", default=None)
current_genclass = contextvars.ContextVar("current_genclass", default=None)
current_cancel_event = contextvars.ContextVar("current_cancel_event", default=None)


def check_cancelled():
    cancel_event = current_cancel_event.get()
    if cancel_event and cancel_event.is_set():
        raise ThreadCancelledError("Task was cancelled")



import asyncio
from contextlib import contextmanager
import os
import sys
import threading
from typing import Optional, Type

from pydantic import BaseModel, Field
from koil.errors import ContextError
from koil.vars import *
from koil.task import KoilGeneratorTask, KoilTask
import time
import logging


logger = logging.getLogger(__name__)

try:
    import uvloop
except:
    uvloop = None


@contextmanager
def _selector_policy(uvify=True):

    original_policy = asyncio.get_event_loop_policy()

    try:
        if uvify:
            if uvloop:
                asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
            else:
                logger.info("uvloop not installed, using default policy")
        elif (
            sys.version_info >= (3, 8)
            and os.name == "nt"
            and hasattr(asyncio, "WindowsSelectorEventLoopPolicy")
        ):
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
        yield
    finally:
        asyncio.set_event_loop_policy(original_policy)


def run_threaded_event_loop(loop):

    try:
        loop.run_forever()
    finally:
        try:
            # mimic asyncio.run() behavior
            # cancel unexhausted async generators
            tasks = asyncio.all_tasks(loop)
            for task in tasks:
                task.cancel()

            async def gather():
                logger.debug(f"Cancelling {tasks}")
                await asyncio.gather(*tasks, return_exceptions=True)

            loop.run_until_complete(gather())
            for task in tasks:
                if task.cancelled():
                    continue
                if task.exception() is not None:
                    loop.call_exception_handler(
                        {
                            "message": "unhandled exception during loop shutdown",
                            "exception": task.exception(),
                            "task": task,
                        }
                    )
            if hasattr(loop, "shutdown_asyncgens"):
                loop.run_until_complete(loop.shutdown_asyncgens())
        finally:
            logger.debug("Loop Sucessfully Closed")
            loop.close()


def get_threaded_loop(name="KoilLoop", uvify=True):
    """Creates a new event loop and run it in a new thread."""
    with _selector_policy(uvify=uvify):
        newloop = asyncio.new_event_loop()

    th = threading.Thread(target=run_threaded_event_loop, args=(newloop,), name=name)
    th.daemon = True
    th.start()

    return newloop


class Koil:
    def __init__(
        self,
        *args,
        name="Koiloop",
        grace_period=None,
        uvify=True,
        task_class: Optional[Type[KoilTask]] = None,
        gen_class: Optional[Type[KoilGeneratorTask]] = None,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.it = "it"
        self.name = name
        self.task_class = task_class
        self.gen_class = gen_class
        self.loop = None
        self.popped_loop = None
        self.uvify = uvify

    async def __aenter__(self):
        self.old_loop = current_loop.get()
        loop = asyncio.get_event_loop()
        current_loop.set(loop)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        current_loop.set(self.old_loop)

    def __enter__(self):
        try:
            asyncio.get_running_loop()
            raise ContextError(
                "You are running in an event loop already. Using koil makes no sense here, use asyncio instead. If this happens in a context manager, you probably forgot to use the `async with` syntax."
            )
        except RuntimeError:
            pass

        self.old_loop = current_loop.get()
        self.old_taskclass = current_taskclass.get()
        self.old_genclass = current_genclass.get()

        current_taskclass.set(
            self.task_class or self.old_taskclass or KoilTask
        )  # task classes can be overwriten, as they only apply to the context
        current_genclass.set(
            self.gen_class or self.old_genclass or KoilGeneratorTask
        )  # task classes can be overwriten, as they only apply to the context
        if self.old_loop is not None:
            # already runnning with a koiled loop, we will just attach to it
            return self

        self.popped_loop = get_threaded_loop(self.name, uvify=self.uvify)
        current_loop.set(self.popped_loop)
        current_koil.set(self)
        return self

    async def aclose(self):
        loop = asyncio.get_event_loop()
        logger.debug("Causing loop to stop")
        loop.stop()

    def __exit__(self, exc_type, exc_val, exc_tb):

        if self.popped_loop is not None:
            asyncio.run_coroutine_threadsafe(self.aclose(), self.popped_loop)

            iterations = 0

            while self.popped_loop.is_running():
                time.sleep(0.001)
                iterations += 1
                if iterations == 100:
                    logger.warning(
                        "Shutting Down takes longer than expected. Probably we are having loose Threads? Keyboard interrupt?"
                    )

            current_loop.set(self.old_loop)  # Reset the loop

        current_taskclass.set(self.old_taskclass)
        current_taskclass.set(self.old_genclass)
        current_koil.set(None)



class KoilError(Exception):
    pass


class KoilStopIteration(Exception):
    """Wrapper around StopIteration as it interacts badly with Futures"""

    pass


class ContextError(KoilError):
    pass


class ThreadCancelledError(KoilError):
    """Mimics ayncio.CancelledError for threads if they accept cancellation. This should only
    be raised from within a thread that was spawned by koil."""

    pass


class CancelledError(KoilError):
    """Mimics ayncio.CancelledError for futures that are living in a koiled loop. Catch this
    as if you would catch asyncio.CancelledError."""

    pass



import asyncio
import contextvars
import concurrent.futures
import threading
from koil.errors import CancelledError


async def check_event(event: threading.Event, timeout: float = None):
    """
    Waits for an event to be set, or for a timeout to occur.
    """
    while True:
        await asyncio.sleep(0.00001)
        if event.is_set():
            return


def run_threaded_with_context(
    future: asyncio.Future,
    loop: asyncio.AbstractEventLoop,
    cancel_event: threading.Event,
) -> concurrent.futures.Future:
    """Runs a future in the supplied loop but copiesthe context
    of the current loop,

    Args:
        future (asyncio.Future): The asyncio FUture
        loop (asyncio.AbstractEventLoop): The loop in which we run?

    Returns:
        concurrent.futures.Future: The future + newcontext
    """

    ctxs = contextvars.copy_context()

    async def passed_with_context(future):

        cancel_f = asyncio.create_task(check_event(cancel_event))
        future_t = asyncio.create_task(future)
        for ctx, value in ctxs.items():
            ctx.set(value)

        finished, unfinished = await asyncio.wait(
            [future_t, cancel_f], return_when=asyncio.FIRST_COMPLETED
        )

        for task in finished:
            if task == cancel_f:
                for untask in unfinished:
                    untask.cancel()
                    try:
                        await untask
                    except asyncio.CancelledError:
                        pass  # we are not interested in this and it should always be fine

                raise CancelledError(f"Future {future} was cancelled")

            for untask in unfinished:
                untask.cancel()
                try:
                    await untask
                except asyncio.CancelledError:
                    pass  # we are not interested in this and it should always be fine

            if task.exception():
                raise task.exception()

            newcontext = contextvars.copy_context()
            return task.result(), newcontext

    return asyncio.run_coroutine_threadsafe(passed_with_context(future), loop)



from koil.helpers import unkoil
from koil.koil import Koil
import inspect
from typing import Callable, Type, TypeVar
from koil.vars import current_loop

T = TypeVar("T")


def koilable(
    fieldname: str = "__koil",
    add_connectors: bool = False,
    koil_class: Type[Koil] = Koil,
    **koilparams
) -> Callable[[Type[T]], Type[T]]:
    """
    Decorator to make an async generator koilable.

    """

    def real_cls_decorator(cls: Type[T]) -> Type[T]:

        cls.__koilable__ = True
        assert hasattr(cls, "__aenter__"), "Class must implement __aenter__"
        assert hasattr(cls, "__aenter__"), "Class must implement __aexit__"
        assert inspect.iscoroutinefunction(
            cls.__aenter__
        ), "__aenter__ must be a coroutine"
        assert inspect.iscoroutinefunction(
            cls.__aexit__
        ), "__aexit__ must be a coroutine"

        def koiled_enter(self, *args, **kwargs):
            if current_loop.get() is None:
                if getattr(self, fieldname, None) is None:
                    setattr(self, fieldname, koil_class(**koilparams))
                getattr(self, fieldname).__enter__()

            return unkoil(self.__aenter__, *args, **kwargs)

        def koiled_exit(self, *args, **kwargs):
            unkoil(self.__aexit__, *args, **kwargs)
            koil = getattr(self, fieldname, None)
            if koil is not None:
                koil.__exit__(None, None, None)
            setattr(self, fieldname, None)

        def disconnect(self):
            unkoil(self.__aexit__, None, None, None)
            koil = getattr(self, fieldname, None)
            if koil is not None:
                koil.__exit__(None, None, None)
            setattr(self, fieldname, None)

        async def adisconnect(self):
            return await self.__aexit__(None, None, None)

        async def aconnect(self):
            return await self.__aenter__()

        cls.__enter__ = koiled_enter
        cls.__exit__ = koiled_exit
        if add_connectors:
            cls.adisconnect = adisconnect
            cls.aconnect = aconnect
            cls.disconnect = disconnect
            cls.connect = koiled_enter

        return cls

    return real_cls_decorator


def unkoilable(func):
    def wrapper(*args, **kwargs):
        return unkoil(func, *args, **kwargs)

    return wrapper



import asyncio
from asyncio.log import logger
import threading

import janus
from koil.errors import KoilError, KoilStopIteration, ThreadCancelledError
from koil.utils import run_threaded_with_context
from koil.vars import *
from koil.task import KoilTask
import time


def create_task(coro, *args, **kwargs) -> KoilTask:
    """Create a coil task from a coroutine"""
    return coro(*args, **kwargs, as_task=True)


def unkoil_gen(iterator, *args, timeout=None, **kwargs):
    loop = current_loop.get()
    cancel_event = current_cancel_event.get() or threading.Event()

    if loop.is_closed():
        raise RuntimeError("Loop is not running")
    try:
        loop0 = asyncio.events.get_running_loop()
        if loop0 is loop:
            raise NotImplementedError("Calling sync() from within a running loop")
    except RuntimeError:
        pass

    ait = iterator(*args, **kwargs).__aiter__()
    res = [False, False]
    next_args = ()

    async def next_on_ait(next_args):
        try:
            try:
                obj = await ait.__anext__(*next_args if next_args else ())
                return [False, obj]
            except StopAsyncIteration:
                return [True, None]
        except asyncio.CancelledError as e:
            return [False, e]

    while True:
        res = run_threaded_with_context(
            next_on_ait(next_args), loop, cancel_event=cancel_event
        )
        x, context = res.result()
        done, obj = x
        if done:
            if obj:
                raise obj
            break

        for ctx, value in context.items():
            ctx.set(value)

        next_args = yield obj


def unkoil(coro, *args, timeout=None, as_task=False, ensure_koiled=False, **kwargs):
    try:
        loop = asyncio.events.get_running_loop()
        if loop == current_loop.get():
            raise NotImplementedError(
                "Calling sync() from within a running loop, you need to await the coroutine"
            )

    except RuntimeError:
        pass

    loop = current_loop.get()
    cancel_event = current_cancel_event.get()

    if loop:
        try:
            if loop.is_closed():
                raise RuntimeError("Loop is not running")

            ctxs = contextvars.copy_context()

            async def passed_with_context():
                for ctx, value in ctxs.items():
                    ctx.set(value)

                x = await coro(*args, **kwargs)
                newcontext = contextvars.copy_context()
                return x, newcontext

            if as_task:
                taskclass = current_taskclass.get()
                assert taskclass is not None, "No task class set"
                return taskclass(
                    coro, preset_args=args, preset_kwargs=kwargs, loop=loop
                )

            co_future = asyncio.run_coroutine_threadsafe(passed_with_context(), loop)
            while not co_future.done():
                time.sleep(0.01)
                if cancel_event and cancel_event.is_set():
                    raise Exception("Task was cancelled")

            x, newcontext = co_future.result()

            for ctx, value in newcontext.items():
                ctx.set(value)

            return x

        except KeyboardInterrupt:
            print("Grace period triggered?")
            raise
    else:
        if ensure_koiled:
            raise RuntimeError("No loop set and ensure_koiled was set to True")

        if as_task:
            raise RuntimeError(
                """No loop is running. That means you cannot have this run as a task. Try providing a loop by entering a Koil() context.
                """
            )

        logger.warn(
            "You used unkoil without a governing Koil in the context, this is not recommended. We will now resort to run asyncio.run()"
        )

        future = coro(*args, **kwargs)
        asyncio.run(future)


async def run_spawned(
    sync_func,
    *sync_args,
    executor=None,
    pass_context=False,
    pass_loop=True,
    cancel_timeout=None,
    **sync_kwargs,
):
    """
    Spawn a thread with a given sync function and arguments
    """

    loop = current_loop.get()
    try:
        loop0 = asyncio.get_event_loop()
        if loop:
            assert loop0 is loop, "Loop is not the same"
        else:
            loop = loop0
            current_taskclass.set(KoilTask)
    except RuntimeError:
        loop = current_loop.get()

    assert loop, "No koiled loop found"
    assert loop.is_running(), "Loop is not running"

    def wrapper(sync_args, sync_kwargs, loop, cancel_event, context):
        if pass_loop:
            current_loop.set(loop)
            current_taskclass.set(KoilTask)

        current_cancel_event.set(cancel_event)

        if context:
            for ctx, value in context.items():
                ctx.set(value)

        print("New thread spawned")
        return sync_func(*sync_args, **sync_kwargs)

    context = contextvars.copy_context() if pass_context else None
    cancel_event = threading.Event()

    f = loop.run_in_executor(
        None, wrapper, sync_args, sync_kwargs, loop, cancel_event, context
    )
    try:
        return await asyncio.shield(f)
    except asyncio.CancelledError as e:
        cancel_event.set()

        try:
            await asyncio.wait_for(f, timeout=cancel_timeout)
        except ThreadCancelledError:
            print("Thread was cancelled")
            logger.info("Future in another thread was sucessfully cancelled")
            pass
        except asyncio.TimeoutError as te:
            raise KoilError(
                f"We could not successfully cancel the future {f} another thread. Make sure you are not blocking the thread with a long running task and check if you check_cancelled every now and then"
            ) from te

        raise e


async def iterate_spawned(
    sync_gen,
    *sync_args,
    executor=None,
    pass_context=False,
    pass_loop=True,
    cancel_timeout=None,
    **sync_kwargs,
):
    """
    Spawn a thread with a given sync function and arguments
    """

    loop = current_loop.get()
    try:
        loop0 = asyncio.get_event_loop()
        if loop:
            assert loop0 is loop, "Loop is not the same"
        else:
            loop = loop0
            current_taskclass.set(KoilTask)
    except RuntimeError:
        loop = current_loop.get()

    assert loop, "No koiled loop found"
    assert loop.is_running(), "Loop is not running"

    yield_queue = janus.Queue()
    next_queue = janus.Queue()
    cancel_event = threading.Event()

    def wrapper(
        sync_args,
        sync_kwargs,
        sync_yield_queue,
        sync_next_queue,
        loop,
        cancel_event,
        context,
    ):
        if pass_loop:
            current_loop.set(loop)
            current_taskclass.set(KoilTask)

        current_cancel_event.set(cancel_event)

        if context:
            for ctx, value in context.items():
                ctx.set(value)

        it = sync_gen(*sync_args, **sync_kwargs).__iter__()

        args = ()
        while True:
            try:
                res = it.__next__(*args if args else ())
                if cancel_event.is_set():
                    raise ThreadCancelledError("Thread was cancelled")
                sync_yield_queue.put(res)
                args = sync_next_queue.get()
                sync_next_queue.task_done()
            except StopIteration as e:
                raise KoilStopIteration("Thread stopped")
            except Exception as e:
                print(e)
                raise e

    context = contextvars.copy_context() if pass_context else None

    f = loop.run_in_executor(
        None,
        wrapper,
        sync_args,
        sync_kwargs,
        yield_queue.sync_q,
        next_queue.sync_q,
        loop,
        cancel_event,
        context,
    )

    while True:

        it_task = asyncio.create_task(yield_queue.async_q.get())

        finish, unfinished = await asyncio.wait(
            [it_task, f], return_when=asyncio.FIRST_COMPLETED
        )

        finish_condition = False

        for task in finish:
            if task == f:
                if task.exception():
                    finish_condition = task.exception()

            else:
                yield_queue.async_q.task_done()
                x = yield task.result()
                await next_queue.async_q.put(x)

        if finish_condition:
            yield_queue.close()
            try:
                raise finish_condition
            except KoilStopIteration:
                break



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['koil', 'koil.checker', 'koil.checker.defaults']

package_data = \
{'': ['*']}

install_requires = \
['janus>=1.0.0,<2.0.0']

extras_require = \
{'jupyter': ['jupyter>=1.0.0,<2.0.0'], 'uvloop': ['uvloop>=0.16.0,<0.17.0']}

setup_kwargs = {
    'name': 'koil',
    'version': '0.1.106',
    'description': 'Async for a sync world',
    'long_description': '# koil\n\n[![codecov](https://codecov.io/gh/jhnnsrs/koil/branch/master/graph/badge.svg?token=UGXEA2THBV)](https://codecov.io/gh/jhnnsrs/koil)\n[![PyPI version](https://badge.fury.io/py/koil.svg)](https://pypi.org/project/koil/)\n![Maintainer](https://img.shields.io/badge/maintainer-jhnnsrs-blue)\n[![PyPI pyversions](https://img.shields.io/pypi/pyversions/koil.svg)](https://pypi.python.org/pypi/koil/)\n[![PyPI status](https://img.shields.io/pypi/status/koil.svg)](https://pypi.python.org/pypi/koil/)\n[![PyPI download day](https://img.shields.io/pypi/dm/koil.svg)](https://pypi.python.org/pypi/koil/)\n\n### DEVELOPMENT\n\n# Quick Start\n\nLet\'s discover **Koil in less than 5 minutes**.\n\n### Inspiration\n\nkoil is an abstraction layer for threaded asyncio to enable "sensible defaults" for\nprogrammers working with frameworks that are barely compatible with asyncio (originally developped to get around pyqt5)\n\n### Main Concept\n\nAsync libraries are amazing, and its an ecosystem rapidly increasing, however in some contexts it still doesn\'t seem like\nthe way to go and the burden of learning these concepts might be to high. However you developed a wonderful async api\nthat you want to share with the world.\n\n```python\nclass AmazingAsyncAPI:\n    def __init__(self) -> None:\n        pass\n\n    async def sleep(self):\n        await asyncio.sleep(0.01)\n        return "the-glory-of-async"\n\n    async def __aenter__(self):\n        # amazing connection logic\n        return self\n\n    async def __aexit__(self, *args, **kwargs):\n        # amazing tear down logic\n        return self\n\n```\n\nHowever if somebody wants to use this api in sync environment they are in for a good one, as a call to asyncio.run() just wont do the trick.\n\n```python\nfrom koil import koilable, unkoilable\n\n@koilable()\nclass AmazingAsyncAPI:\n    def __init__(self) -> None:\n        pass\n\n    @unkoilable()\n    async def sleep(self):\n        await asyncio.sleep(0.01)\n        return "the-glory-of-async"\n\n    async def __aenter__(self):\n        # amazing connection logic\n        return self\n\n    async def __aexit__(self, *args, **kwargs):\n        # amazing tear down logic\n        return self\n\n```\n\nAnd now it works. Just use your Api with a normal context manager.\n\n```python\nwith AmazingAsyncAPI as e:\n  print(e.sleep())\n```\n\nKoil under the hood spawns a new event loop in another thread, calls functions that are marked with unkoilable\nthreadsafe in that loop and returns the result, when exiting it shuts down the loop in the other thread.\n\nIf you have multiple context managers or tasks that you would just like to run in another thread, you can\nalso create a loop in another thread\n\n```python\n\nasync def task(arg):\n       x = await ...\n       return x\n\nwith Koil(): # creates a threaded loop\n\n    x = unkoil(task, 1)\n\n    with AmazingAsyncAPI as e:\n       print(e.sleep())\n\n```\n\nMoreover koil also is able to be used with generators\n\n```python\nimport asyncio\nfrom koil import unkoil_gen\n\nasync def task(arg):\n    for i in range(0,20)\n      await asyncio.sleep(1)\n      yield arg\n\n\nwith Koil(): # creates a threaded loop\n\n    for x in unkoil_gen(task, 1):\n        print(x)\n\n```\n\nAnd finally koil is able to create task like objects,\n\n```python\nasync def task(arg):\n    await asyncio.sleep(2)\n    return arg\n\nwith Koil():\n\n  x = unkoil(task, 1, as_task=True)\n\n  # do other stuff\n\n  if x.done():\n      print(x)\n\n```\n\n## PyQt Support\n\n... Documentation coming soon...\n\n### Installation\n\n```bash\npip install koil\n```\n',
    'author': 'jhnnsrs',
    'author_email': 'jhnnsrs@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import asyncio
from pydantic import BaseModel, Field, root_validator

from koil.decorators import koilable
from typing import Optional, Type, TypeVar

from koil.koil import Koil
from koil.qt import QtGeneratorTask, QtTask
from koil.task import KoilGeneratorTask, KoilTask
from koil.vars import *
from koil.koil import *
from qtpy import QtWidgets, QtCore

T = TypeVar("T")


class PedanticKoil(BaseModel):
    name: str = "KoilLoop"
    uvify: bool = True
    grace_period: Optional[float] = None
    task_class: Optional[Type[KoilTask]] = Field(default=KoilTask, exclude=True)
    gen_class: Optional[Type[KoilGeneratorTask]] = Field(
        default=KoilGeneratorTask, exclude=True
    )
    grant_sync = True

    _entered_loop: asyncio.BaseEventLoop = None
    _old_loop: asyncio.BaseEventLoop = None
    _old_taskclass: Type[KoilTask] = None
    _old_genclass: Type[KoilGeneratorTask] = None

    @root_validator()
    def check_not_running_in_loop(cls, values):
        if current_loop.get() is not None:
            raise ValueError(
                "You are already running in a Koil Loop. You cannot run a Koil Loop inside another Koil Loop."
            )
        try:
            asyncio.get_running_loop()
            if not values["grant_sync"]:
                raise ValueError(
                    "Please use async instead. Or set Koil to grant_sync=True"
                )
        except RuntimeError:
            pass

        return values

    def connect(self):
        return self.__enter__()

    async def aconnect(self):
        return self.__aenter__()

    def disconnect(self):
        return self.__exit__(None, None, None)

    async def adisconnect(self):
        return self.__aexit__(None, None, None)

    async def __aenter__(self):
        self._old_loop = current_loop.get()
        loop = asyncio.get_event_loop()
        current_loop.set(loop)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        current_loop.set(self._old_loop)

    def __enter__(self):
        try:
            asyncio.get_running_loop()
            if not self.grant_sync:
                raise ContextError(
                    "You are running in an event loop already. Using koil makes no sense here, use asyncio instead. If this happens in a context manager, you probably forgot to use the `async with` syntax."
                )
        except RuntimeError:
            pass

        self._old_loop = current_loop.get()
        self._old_taskclass = current_taskclass.get()
        self._old_genclass = current_genclass.get()

        current_taskclass.set(
            self.task_class or self._old_taskclass or KoilTask
        )  # task classes can be overwriten, as they only apply to the context
        current_genclass.set(
            self.gen_class or self._old_genclass or KoilGeneratorTask
        )  # task classes can be overwriten, as they only apply to the context
        if self._old_loop is not None:
            # already runnning with a koiled loop, we will just attach to it
            return self

        self._entered_loop = get_threaded_loop(self.name, uvify=self.uvify)
        current_loop.set(self._entered_loop)
        return self

    async def aclose(self):
        loop = asyncio.get_event_loop()
        logger.debug("Causing loop to stop")
        loop.stop()

    def __exit__(self, exc_type, exc_val, exc_tb):

        if self._entered_loop is not None:
            if self._entered_loop.is_running():
                asyncio.run_coroutine_threadsafe(self.aclose(), self._entered_loop)

                iterations = 0

                while self._entered_loop.is_running():
                    time.sleep(0.001)
                    iterations += 1
                    if iterations == 100:
                        logger.warning(
                            "Shutting Down takes longer than expected. Probably we are having loose Threads? Keyboard interrupt?"
                        )

                current_loop.set(self._old_loop)  # Reset the loop

        current_taskclass.set(self._old_taskclass)
        current_taskclass.set(self._old_genclass)

    class Config:
        arbitraty_types_allowed = True
        underscore_attrs_are_private = True


@koilable(fieldname="koil", add_connectors=True, koil_class=PedanticKoil)
class KoiledModel(BaseModel):
    koil: Optional[PedanticKoil]

    def __enter__(self: T) -> T:
        ...

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        ...

    async def __aenter__(self: T) -> T:
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    class Config:
        arbitrary_types_allowed = True


class Composition(KoiledModel):
    async def __aenter__(self: T) -> T:
        for key, value in self:
            if isinstance(value, PedanticKoil):
                continue  # that was entered before
            if hasattr(value, "__aenter__"):
                await value.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        for key, value in self:
            if isinstance(value, PedanticKoil):
                continue  # that was entered before
            if hasattr(value, "__aexit__"):
                await value.__aexit__(exc_type, exc_val, exc_tb)

    def _repr_html_(self):
        return (
            "<div><p>App</p><table>"
            + "\n".join(["<tr><td>{}</td></tr>".format(key) for key, value in self])
            + "</table></div>"
        )


class WrappedObject(QtCore.QObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def set_koil(self, koil: PedanticKoil):
        self.koil = koil

    def close(self):
        self.koil.__exit__(None, None, None)


class QtPedanticKoil(PedanticKoil):
    disconnect_on_close = True
    auto_connect = True
    task_class: Optional[Type[QtTask]] = Field(default=QtTask, exclude=True)
    gen_class: Optional[Type[QtGeneratorTask]] = Field(
        default=QtGeneratorTask, exclude=True
    )
    qobject: WrappedObject = Field(default_factory=WrappedObject)

    def __init__(self, **data):
        super().__init__(**data)
        if self.auto_connect:
            self.__enter__()

    def __enter__(self):
        self.qobject.set_koil(self)
        ap_instance = QtWidgets.QApplication.instance()
        if ap_instance is None:
            raise NotImplementedError("Qt Application not found")
        if self.disconnect_on_close:
            ap_instance.lastWindowClosed.connect(self.qobject.close)
        return super().__enter__()

    class Config:
        arbitrary_types_allowed = True
